//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/Sample/SamplesQListView.cpp
//! @brief     Implements class SamplesQListView.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/Sample/SamplesQListView.h"
#include "Base/Util/SysUtil.h"
#include "GUI/Model/FromCore/GUIExamplesFactory.h"
#include "GUI/Model/Project/ProjectDocument.h"
#include "GUI/Model/Sample/SampleItem.h"
#include "GUI/Model/Sample/SamplesSet.h"
#include "GUI/View/Layout/ApplicationSettings.h"
#include "GUI/View/Sample/SamplesQModel.h"
#include "GUI/View/Tool/ItemDelegateForHTML.h"
#include "GUI/View/Widget/ItemViewOverlayButtons.h"
#include <QAction>
#include <QMenu>
#include <QPainter>

namespace {

class ItemDelegateForSampleTree : public ItemDelegateForHTML {
public:
    ItemDelegateForSampleTree(QObject* parent)
        : ItemDelegateForHTML(parent)
    {
    }

protected:
    void paint(QPainter* painter, const QStyleOptionViewItem& option,
               const QModelIndex& index) const override
    {
        ItemDelegateForHTML::paint(painter, option, index);

        QStyleOptionViewItem options = option;
        initStyleOption(&options, index);

        painter->save();

        painter->setPen(QPen(Qt::lightGray, 1));
        painter->drawLine(options.rect.left(), options.rect.bottom(), options.rect.right(),
                          options.rect.bottom());
        painter->restore();
    }

    QSize sizeHint(const QStyleOptionViewItem& option, const QModelIndex& index) const override
    {
        auto s = ItemDelegateForHTML::sizeHint(option, index);
        s.setHeight(std::max(s.height(), 32));
        return s;
    }
};

} // namespace


SamplesQListView::SamplesQListView()
    : m_model(new SamplesQModel)
    , m_new_sample_action(new QAction(this))
    , m_import_sample_action(new QAction(this))
    , m_choose_from_library_action(new QAction(this))
{
    setMaximumWidth(200);

    setContextMenuPolicy(Qt::CustomContextMenu);
    setModel(m_model);
    setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);

    ItemViewOverlayButtons::install(
        this, [this](const QModelIndex& i, bool h) { return getOverlayActions(i, h); });
    setItemDelegate(new ItemDelegateForSampleTree(this));

    connect(selectionModel(), &QItemSelectionModel::currentChanged, this,
            &SamplesQListView::onCurrentChanged);

    connect(this, &QWidget::customContextMenuRequested, this, &SamplesQListView::showContextMenu);


    m_new_sample_action->setText("Create new sample");
    m_new_sample_action->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_new_sample_action->setIconText("New");
    m_new_sample_action->setToolTip("Create new sample");
    connect(m_new_sample_action, &QAction::triggered, this, &SamplesQListView::createNewSample);


    m_import_sample_action->setText("Import sample from python script (experimental)");
    m_import_sample_action->setIcon(QIcon(":/images/import.svg"));
    m_import_sample_action->setIconText("Import");
    m_import_sample_action->setToolTip(
        "Import sample from Python script.\n The script should contain a function "
        "returning a valid multi-layer.");

#ifdef BORNAGAIN_PYTHON
    connect(m_import_sample_action, &QAction::triggered, this,
            &SamplesQListView::importSampleFromPython);
#else
    m_import_sample_action->setVisible(false);
#endif


    m_choose_from_library_action->setText("Choose from sample examples");
    m_choose_from_library_action->setIcon(QIcon(":/images/library.svg"));
    m_choose_from_library_action->setIconText("Examples");
    m_choose_from_library_action->setToolTip("Choose from sample examples");

    auto* menu = new QMenu(this);
    m_choose_from_library_action->setMenu(menu);

    for (const auto& exampleName : GUI::ExamplesFactory::exampleNames()) {
        QString title, description;
        std::tie(title, description) = GUI::ExamplesFactory::exampleInfo(exampleName);
        auto icon = QIcon(":/images/sample_layers2.png");
        auto* action = menu->addAction(icon, title);
        action->setToolTip(description);
        connect(action, &QAction::triggered, [this, exampleName, title, description] {
            createSampleFromLibrary(exampleName, title, description);
        });
    }
}

void SamplesQListView::setCurrentSample(SampleItem* sample)
{
    setCurrentIndex(m_model->indexForItem(sample));
}

SampleItem* SamplesQListView::currentSampleItem()
{
    return m_model->itemForIndex(currentIndex());
}
QSize SamplesQListView::sizeHint() const
{
    QSize s = QListView::sizeHint();
    s.setWidth(std::max(300, s.width()));
    return s;
}

void SamplesQListView::onSampleModified()
{
    // if length of description changes height of TreeItem, the position of the overlay buttons
    // has to be updated -> schedule a re-layout
    scheduleDelayedItemsLayout();
}

void SamplesQListView::createNewSample()
{
    const QModelIndex newIndex = m_model->createSample();
    setCurrentIndex(newIndex);
    gDoc->setModified();
}

void SamplesQListView::createSampleFromLibrary(const QString& classname, const QString& title,
                                               const QString& description)
{
    const QModelIndex newIndex = m_model->createSampleFromExamples(classname, title, description);
    if (newIndex.isValid()) {
        setCurrentIndex(newIndex);
        gDoc->setModified();
    }
}

void SamplesQListView::importSampleFromPython()
{
#ifdef BORNAGAIN_PYTHON
    const QModelIndex newIndex = m_model->createSampleFromPython();
    if (newIndex.isValid()) {
        setCurrentIndex(newIndex);
        gDoc->setModified();
    }
#endif
}

QList<QAction*> SamplesQListView::getOverlayActions(const QModelIndex& index, bool asHover)
{
    if (!asHover)
        return {};

    auto* item = m_model->itemForIndex(index);
    if (item == nullptr)
        return {};

    return {createRemoveAction(this, item)};
}

void SamplesQListView::onCurrentChanged(const QModelIndex& index)
{
    gDoc->samplesModifier()->setCurrentIndex(index.row());
    emit currentSampleChanged(m_model->itemForIndex(index));
}

void SamplesQListView::restoreSelection()
{
    int selected = gDoc->samples()->currentIndex();
    if (selected < 0 || selected >= m_model->rowCount())
        selected = m_model->rowCount() - 1;

    QModelIndex currentIndex = m_model->index(selected, 0, QModelIndex());
    if (currentIndex.isValid()) {
        selectionModel()->select(currentIndex, QItemSelectionModel::ClearAndSelect);
        setCurrentSample(gDoc->samples()->sampleItems()[selected]);
    }
}

void SamplesQListView::removeSample(SampleItem* item)
{
    m_model->removeSample(item);
    gDoc->setModified();
}

void SamplesQListView::copySample(SampleItem* item)
{
    m_model->copySample(item);
    gDoc->setModified();
}

QAction* SamplesQListView::createRemoveAction(QObject* parent, SampleItem* item)
{
    auto* removeAction = new QAction(parent);
    removeAction->setText("Remove");
    removeAction->setIcon(QIcon(":/images/delete.svg"));
    removeAction->setIconText("Remove");
    removeAction->setToolTip("Remove this sample");
    connect(removeAction, &QAction::triggered, [this, item] { removeSample(item); });

    return removeAction;
}

QAction* SamplesQListView::createCopyAction(QObject* parent, SampleItem* item)
{
    auto* copyAction = new QAction(parent);
    copyAction->setText("Clone");
    copyAction->setIcon(QIcon(":/images/content-copy.svg"));
    copyAction->setIconText("Clone");
    copyAction->setToolTip("Clone this sample");
    connect(copyAction, &QAction::triggered, [this, item] { copySample(item); });

    return copyAction;
}

void SamplesQListView::showContextMenu(const QPoint& pos)
{
    auto* sampleAtPoint = m_model->itemForIndex(indexAt(pos));

    QMenu menu;
    menu.setToolTipsVisible(true);

    if (sampleAtPoint != nullptr) {
        menu.addAction(createCopyAction(&menu, sampleAtPoint));
        menu.addAction(createRemoveAction(&menu, sampleAtPoint));
        menu.addSeparator();
    }

    menu.addAction(m_new_sample_action);
    menu.addAction(m_choose_from_library_action);
    menu.exec(mapToGlobal(pos));
}
